#ifndef SUBPAR_SUBPAR_HPP
#define SUBPAR_SUBPAR_HPP

#include "range.hpp"
#include "simple.hpp"

/**
 * @file subpar.hpp
 * @brief Substitutable parallelization functions.
 */

/**
 * @namespace subpar
 * @brief Substitutable parallelization functions.
 */
namespace subpar {}

#endif
