#ifndef KNNCOLLE_HPP
#define KNNCOLLE_HPP

#include "Builder.hpp"
#include "Prebuilt.hpp"
#include "MockMatrix.hpp"

#include "distances.hpp"
#include "L2Normalized.hpp"

#include "Bruteforce.hpp"
#include "Vptree.hpp"
#include "Kmknn.hpp"
#include "find_nearest_neighbors.hpp"

/**
 * @file knncolle.hpp
 *
 * @brief Umbrella header for all algorithms.
 */

/**
 * @brief Collection of KNN algorithms.
 * @namespace knncolle
 */
namespace knncolle {}

#endif

