#ifndef TATAMI_TATAMI_STATS_HPP
#define TATAMI_TATAMI_STATS_HPP

#include "counts.hpp"
#include "grouped_medians.hpp"
#include "grouped_sums.hpp"
#include "grouped_variances.hpp"
#include "medians.hpp"
#include "ranges.hpp"
#include "sums.hpp"
#include "utils.hpp"
#include "variances.hpp"

/**
 * @file tatami_stats.hpp
 * @brief Umbrella header for the **tatami_stats** library.
 */

/**
 * @namespace tatami_stats
 * @brief Functions to compute statistics from a `tatami::Matrix`.
 */
namespace tatami_stats {}

#endif
