#ifndef SINGLEPP_UMBRELLA_HPP
#define SINGLEPP_UMBRELLA_HPP

/**
 * @file singlepp.hpp
 *
 * @brief Umbrella header for the **singlepp** library.
 */

#include "Markers.hpp"
#include "Intersection.hpp"
#include "choose_classic_markers.hpp"
#include "train_single.hpp"
#include "train_integrated.hpp"
#include "classify_single.hpp"
#include "classify_integrated.hpp"

/**
 * @namespace singlepp
 * @brief Cell type classification using the SingleR algorithm in C++.
 */
namespace singlepp {}

#endif
